/*
  Token: final .
  NodeOptional: final_if_present.
  NodeSequence: separating_sequence, final.
  NodeList(Optional): separating_list, final.
*/



//
// Pretty printer for the Java grammar.
// Author: Kevin Tao, taokr@cs
//
// VERSION: 2c
//
// Newly added:
//    2c  - 9/21/97: Modified for use with JTB 0.9.0b; visitXXX() methods
//          changed to visit() to use overloading
//    2b  - Minor cosmetic changes, changed visibility of "spc" to protected.
//    2a  - Corrected errors in WhileStatement, and Break and ContinueStatement.
//    2a  - Corrected a few slight formatting errors.
//    2   - flushWriter() method to flush the output stream.
//
// (reminders for myself):
// - spc.spc should be printed after every "\n" or println().
// - println() should not be the last thing printed in a visit method.
//

/* RCS Information
 *
 * $Id: PrettyPrinter.java,v 1.2 1997/10/25 20:41:21 ktao Exp $
 *
 * $Log: PrettyPrinter.java,v $
 * Revision 1.2  1997/10/25 20:41:21  ktao
 * - Added log info.
 * - Deleted irrelevant comment.
 *
 */

package visitor;

//import misc.Spacing;
import syntaxtree.*;
import java.util.*;
import java.io.*;

public class PrettyPrinter extends Visitor {
    protected int anonymousClassLevel = 0;
   protected PrintWriter out;
   protected Spacing spc = new Spacing(3);   // see below for class Spacing
    protected String sep;
    protected String finalSep;
    
   public PrettyPrinter()       { out = new PrintWriter(System.out, true); }
   public PrettyPrinter(Writer o)        { out = new PrintWriter(o, true); }
   public PrettyPrinter(OutputStream o)  { out = new PrintWriter(o, true); }

   public void flushWriter()     { out.flush(); }

   //
   // A few convenience methods overloading visit()
   //
   public void visit(NodeList n) {
      for ( Enumeration e = n.elements(); e.hasMoreElements(); ) {
         ((Node)e.nextElement()).accept(this);
         if ( e.hasMoreElements() ) separate();
      }
      finalSeparate();
   }

   public void visit(NodeListOptional n) {
       if ( n.present() ) {
         for ( Enumeration e = n.elements(); e.hasMoreElements(); ) {
            ((Node)e.nextElement()).accept(this);
            if ( e.hasMoreElements() ) separate();
         }
	 finalSeparate();
       }
   }

   public void visit(NodeOptional n) {
      if ( n.present() ) {
         n.node.accept(this);
         finalSeparate();
      }
   }

   //
   // For convenience, trusts that the node passed to it is a NodeSequence.
   // (of course, will throw an exception if it isn't).
   //
   public void visit(NodeSequence n) {
      for ( Enumeration e = n.elements(); e.hasMoreElements(); ) {
         ((Node)e.nextElement()).accept(this);
         if ( e.hasMoreElements() ) separate();
      }
      finalSeparate();
   }

    
    public void visit(Token n) {
	String currentTokenAsString;

	for (Enumeration e = n.getSpecialTokens(); e.hasMoreElements();) {
	    currentTokenAsString = e.nextElement().toString();
	    out.print(currentTokenAsString);
	    if (currentTokenAsString.endsWith("*/")) {
		newLine();
		out.print(spc.spc);
	    } else if (currentTokenAsString.endsWith("\n")) {
		out.print(spc.spc);
	    }
	}
	out.print(n.toString());
	separate();
    }

    
    public void separate() { out.print(sep); }
    public void finalSeparate() { out.print(finalSep); }

    
    public void setSep(String s) { sep = s; }
    public void setFinalSep(String s) { finalSep = s; }
    
   //
   // Auto class visitors--probably don't need to be overridden.
   //

   //
   // User-generated visitor methods below
   //


    protected void space() { out.print(" "); }

    protected void newLine() { out.println(); }

    protected void updateAnonymousClassLevel(int increment) {
	anonymousClassLevel += increment;
    }
    
    
   //
   // f0 -> [ PackageDeclaration() ]
   // f1 -> ( ImportDeclaration() )*
   // f2 -> ( TypeDeclaration() )*
   // f3 -> < EOF >
   //
   public void visit(CompilationUnit n) {
      out.print(spc.spc);

      setFinalSep("\n\n" + spc.spc);
      n.f0.accept(this);

      setSep("\n" + spc.spc);
      setFinalSep("\n\n" + spc.spc);
      n.f1.accept(this);

      setSep("\n \n \n" + spc.spc);
      setFinalSep("\n");
      n.f2.accept(this);
   }

   //
   // f0 -> "package"
   // f1 -> Name()
   // f2 -> ";"
   //
   public void visit(PackageDeclaration n) {
       n.f0.accept(this); space();
       n.f1.accept(this); space();
       n.f2.accept(this);
   }

   //
   // f0 -> "import"
   // f1 -> Name()
   // f2 -> [ "." "*" ]
   // f3 -> ";"
   //
   // uses the toString() method of the Tokens to print manually
   // (rather than visiting the tokens).
   //
   public void visit(ImportDeclaration n) {
       n.f0.accept(this); space();
       n.f1.accept(this);
       clearSep(); clearFinalSep(); n.f2.accept(this);
       n.f3.accept(this);
   }

   //
   // f0 -> ClassDeclaration()
   //       | InterfaceDeclaration()
   //       | ";"
   //
   public void visit(TypeDeclaration n) {
      n.f0.accept(this);
   }

   //
   // f0 -> ( "abstract"| "final" | "public"  )*
   // f1 -> UnmodifiedClassDeclaration()
   //
   public void visit(ClassDeclaration n) {
       setSep(" ");
       n.f0.accept(this);
       n.f1.accept(this);
   }

   //
   // f0 -> "class"
   // f1 -> < IDENTIFIER >
   // f2 -> [ "extends" Name() ]
   // f3 -> [ "implements" NameList() ]
   // f4 -> ClassBody()
   //
   // continued from ClassDeclaration
   //
   public void visit(UnmodifiedClassDeclaration n) {
       n.f0.accept(this, FINAL_SPACE); 
       n.f1.accept(this, FINAL_SPACE); 

       n.f2.accept(this, SEPARATING_SPACE_FINAL_SPACE);
       n.f3.accept(this, SEPARATING_SPACE_FINAL_SPACE);

       n.f4.accept(this);
   }

   //
   // f0 -> "{"
   // f1 -> ( ClassBodyDeclaration() )*
   // f2 -> "}"
   //
   public void visit(ClassBody n) {
       n.f0.accept(this); newLine();
       if (anonymousClassLevel == 0)
	  newLine();
      if ( n.f1.present() ) {
         spc.updateSpc(+1);
         out.print(spc.spc);
	 setFinalSep((anonymousClassLevel == 0 ? "\n\n" : "\n") + spc.spc);
         n.f1.accept(this);
	 newLine();
         spc.updateSpc(-1);
      }
      if (anonymousClassLevel == 0)
	  newLine();
      out.print(spc.spc);
      n.f2.accept(this);
   }

   //
   // f0 -> ("static"|"abstract"|"final"|"public"|"protected"|"private")*
   // f1 -> UnmodifiedClassDeclaration()
   //
   public void visit(NestedClassDeclaration n) {
      if (n.f0.present()) {
	  visit(n.f0, " ");
	  space();
      }
      n.f1.accept(this);
   }

   //
   // f0 -> Initializer()
   //       | NestedClassDeclaration()
   //       | NestedInterfaceDeclaration()
   //       | ConstructorDeclaration()
   //       | MethodDeclaration()
   //       | FieldDeclaration()
   //
   public void visit(ClassBodyDeclaration n) {
      n.f0.accept(this);
   }

   //
   // f0 -> ( "public"| "protected" | "private" | "static" | "abstract" | "final" | "native" | "synchronized"  )*
   // f1 -> ResultType()
   // f2 -> < IDENTIFIER >
   // f3 -> "("
   //
   // For lookahead only, so don't need to do this
   public void visit(MethodDeclarationLookahead n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      n.f3.accept(this);
   }

   //
   // f0 -> ( "abstract"| "public"  )*
   // f1 -> UnmodifiedInterfaceDeclaration()
   //
   public void visit(InterfaceDeclaration n) {
      visit(n.f0, " ");
      if (n.f0.present())
	  space();
      n.f1.accept(this);
   }

   //
   // f0 -> ("static"|"abstract"|"final"|"public"|"protected"|"private")*
   // f1 -> UnmodifiedInterfaceDeclaration()
   //
   public void visit(NestedInterfaceDeclaration n) {
      visit(n.f0, " ");
      space();
      n.f1.accept(this);
   }

   //
   // f0 -> "interface"
   // f1 -> < IDENTIFIER >
   // f2 -> [ "extends" NameList() ]
   // f3 -> "{"
   // f4 -> ( InterfaceMemberDeclaration() )*
   // f5 -> "}"
   //
   public void visit(UnmodifiedInterfaceDeclaration n) {
       visit(n.f0); space();
       visit(n.f1); space();
      if ( n.f2.present() ) {
	  visit((NodeSequence) n.f2.node, " ");
      }
      visit(n.f3); newLine();
      spc.updateSpc(+1);
      out.print(spc.spc);
      visit(n.f4, "\n" + spc.spc);
      spc.updateSpc(-1);
      out.print("\n" + spc.spc);
      visit(n.f5);
   }

   //
   // f0 -> NestedClassDeclaration()
   //       | NestedInterfaceDeclaration()
   //       | MethodDeclaration()
   //       | FieldDeclaration()
   //
   public void visit(InterfaceMemberDeclaration n) {
      n.f0.accept(this);
   }

   //
   // f0 -> ("public"|"protected"|"private"|"static"|"final"|"transient"|"volatile")*
   // f1 -> Type()
   // f2 -> VariableDeclarator()
   // f3 -> ( "," VariableDeclarator() )*
   // f4 -> ";"
   //
   public void visit(FieldDeclaration n) {
      if ( n.f0.present() ) { visit(n.f0, " "); space(); }
      n.f1.accept(this);
      space();
      n.f2.accept(this);
      n.f3.accept(this);
      visit(n.f4);
   }

   //
   // f0 -> VariableDeclaratorId()
   // f1 -> [ "=" VariableInitializer() ]
   //
   public void visit(VariableDeclarator n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         space();
         visit((NodeSequence) n.f1.node, " ");
      }
   }

   //
   // f0 -> < IDENTIFIER >
   // f1 -> ( "[" "]" )*
   //
   public void visit(VariableDeclaratorId n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   //
   // f0 -> ArrayInitializer()
   //       | Expression()
   //
   public void visit(VariableInitializer n) {
      n.f0.accept(this);
   }

   //
   // f0 -> "{"
   // f1 -> [ VariableInitializer() ( "," VariableInitializer() )* ]
   // f2 -> [ "," ]
   // f3 -> "}"
   //
   public void visit(ArrayInitializer n) {
      visit(n.f0);
      if (n.f1.present()) {
	  NodeSequence seq = (NodeSequence) n.f1.node;
	  seq.elementAt(0).accept(this);
	  visitCommaSeparatedList(((NodeListOptional) seq.elementAt(1))
				  .elements());
      }
      n.f2.accept(this);
      visit(n.f3);
   }

   //
   // f0 -> ( "public"| "protected" | "private" | "static" | "abstract" | "final" | "native" | "synchronized"  )*
   // f1 -> ResultType()
   // f2 -> MethodDeclarator()
   // f3 -> [ "throws" NameList() ]
   // f4 -> ( Block()| ";"  )
   //
   public void visit(MethodDeclaration n) {
      if ( n.f0.present() ) {
         visit(n.f0, " ");
         space();
      }
      n.f1.accept(this);
      space();
      n.f2.accept(this);
      if ( n.f3.present() ) {
         space();
         visit((NodeSequence) n.f3.node, " ");
      }
      space(); 
      n.f4.accept(this);
   }

   //
   // f0 -> < IDENTIFIER >
   // f1 -> FormalParameters()
   // f2 -> ( "[" "]" )*
   //
   public void visit(MethodDeclarator n) {
      visit(n.f0);
      n.f1.accept(this);
      n.f2.accept(this);
   }

   //
   // f0 -> "("
   // f1 -> [ FormalParameter() ( "," FormalParameter() )* ]
   // f2 -> ")"
   //
   public void visit(FormalParameters n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
   }

   //
   // f0 -> [ "final" ]
   // f1 -> Type()
   // f2 -> VariableDeclaratorId()
   //
   public void visit(FormalParameter n) {
      if ( n.f0.present() ) visit(n.f0.node, " ");
      n.f1.accept(this);
      space();
      n.f2.accept(this);
   }

   //
   // f0 -> [ "public"| "protected" | "private"  ]
   // f1 -> < IDENTIFIER >
   // f2 -> FormalParameters()
   // f3 -> [ "throws" NameList() ]
   // f4 -> "{"
   // f5 -> [ ExplicitConstructorInvocation() ]
   // f6 -> ( BlockStatement() )*
   // f7 -> "}"
   //
   public void visit(ConstructorDeclaration n) {
      if ( n.f0.present() ) {
         n.f0.accept(this);
         space();
      }
      visit(n.f1);
      n.f2.accept(this);
      if ( n.f3.present() ) { visit(n.f3.node, " "); }
      space();
      visit(n.f4);
      newLine();
      spc.updateSpc(+1);
      out.print(spc.spc);
      if ( n.f5.present() ) {
	  n.f5.accept(this);
	  out.print("\n" + spc.spc);
      }
      visit(n.f6, "\n" + spc.spc);
      spc.updateSpc(-1);
      out.print("\n" + spc.spc);
      n.f7.accept(this);
   }

   //
   // f0 -> "this" Arguments() ";"
   //       | [ PrimaryExpression() "." ] "super" Arguments() ";"
   //
   public void visit(ExplicitConstructorInvocation n) {
      //
      // This code may not be appropriate if "which" is removed from NodeChoice
      //
      if ( n.f0.which == 0 ) {
         n.f0.accept(this);
      }
      else {
         NodeSequence seq = (NodeSequence)n.f0.choice;
         if (((NodeOptional) seq.elementAt(0)).present()) {
	     seq.elementAt(0).accept(this);
	     space();
	 }
         seq.elementAt(1).accept(this);
         seq.elementAt(2).accept(this);
         seq.elementAt(3).accept(this);
      }
   }

   //
   // f0 -> [ "static" ]
   // f1 -> Block()
   //
   public void visit(Initializer n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   //
   // f0 -> ( PrimitiveType()| Name()  )
   // f1 -> ( "[" "]" )*
   //
   public void visit(Type n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   //
   // f0 -> "boolean"
   //       | "char"
   //       | "byte"
   //       | "short"
   //       | "int"
   //       | "long"
   //       | "float"
   //       | "double"
   //
   public void visit(PrimitiveType n) {
      n.f0.accept(this);
   }

   //
   // f0 -> "public void"
   //       | Type()
   //
   public void visit(ResultType n) {
      n.f0.accept(this);
   }

   //
   // f0 -> < IDENTIFIER >
   // f1 -> ( "." < IDENTIFIER >  )*
   //
   public void visit(Name n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   //
   // f0 -> Name()
   // f1 -> ( "," Name() )*
   //
   // Add spaces?
   //
   public void visit(NameList n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   //
   // f0 -> Assignment()
   //       | ConditionalExpression()
   //
   public void visit(Expression n) {
      n.f0.accept(this);
   }

   //
   // f0 -> PrimaryExpression()
   // f1 -> AssignmentOperator()
   // f2 -> Expression()
   //
   public void visit(Assignment n) {
      n.f0.accept(this);
      space();
      n.f1.accept(this);
      space();
      n.f2.accept(this);
   }

   //
   // f0 -> "="
   //       | "*="
   //       | "/="
   //       | "%="
   //       | "+="
   //       | "-="
   //       | "<<="
   //       | ">>="
   //       | ">>>="
   //       | "&="
   //       | "^="
   //       | "|="
   //
   public void visit(AssignmentOperator n) {
      n.f0.accept(this);
   }

   //
   // f0 -> ConditionalOrExpression()
   // f1 -> [ "?" Expression() ":" ConditionalExpression() ]
   //
   public void visit(ConditionalExpression n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         space();
         visit((NodeSequence) n.f1.node, " ");
      }
   }

   //
   // f0 -> ConditionalAndExpression()
   // f1 -> ( "||" ConditionalAndExpression() )*
   //
   public void visit(ConditionalOrExpression n) {
      n.f0.accept(this);
      visitOperatorList(n.f1);
   }

   //
   // f0 -> InclusiveOrExpression()
   // f1 -> ( "&&" InclusiveOrExpression() )*
   //
   public void visit(ConditionalAndExpression n) {
      n.f0.accept(this);
      visitOperatorList(n.f1);
   }

   //
   // f0 -> ExclusiveOrExpression()
   // f1 -> ( "|" ExclusiveOrExpression() )*
   //
   public void visit(InclusiveOrExpression n) {
      n.f0.accept(this);
      visitOperatorList(n.f1);
   }

   //
   // f0 -> AndExpression()
   // f1 -> ( "^" AndExpression() )*
   //
   public void visit(ExclusiveOrExpression n) {
      n.f0.accept(this);
      visitOperatorList(n.f1);
   }

   //
   // f0 -> EqualityExpression()
   // f1 -> ( "&" EqualityExpression() )*
   //
   public void visit(AndExpression n) {
      n.f0.accept(this);
      visitOperatorList(n.f1);
   }

   //
   // f0 -> InstanceOfExpression()
   // f1 -> ( ( "=="| "!="  ) InstanceOfExpression() )*
   //
   public void visit(EqualityExpression n) {
      n.f0.accept(this);
      visitOperatorList(n.f1);
   }

   //
   // f0 -> RelationalExpression()
   // f1 -> [ "instanceof" Type() ]
   //
   public void visit(InstanceOfExpression n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         space();
         visit((NodeSequence) n.f1.node, " ");
      }
   }

   //
   // f0 -> ShiftExpression()
   // f1 -> ( ( "<"| ">" | "<=" | ">="  ) ShiftExpression() )*
   //
   public void visit(RelationalExpression n) {
      n.f0.accept(this);
      visitOperatorList(n.f1);
   }

   //
   // f0 -> AdditiveExpression()
   // f1 -> ( ( "<<"| ">>" | ">>>"  ) AdditiveExpression() )*
   //
   public void visit(ShiftExpression n) {
      n.f0.accept(this);
      visitOperatorList(n.f1);
   }

   //
   // f0 -> MultiplicativeExpression()
   // f1 -> ( ( "+"| "-"  ) MultiplicativeExpression() )*
   //
   public void visit(AdditiveExpression n) {
      n.f0.accept(this);
      visitOperatorList(n.f1);
   }

   //
   // f0 -> UnaryExpression()
   // f1 -> ( ( "*"| "/" | "%"  ) UnaryExpression() )*
   //
   public void visit(MultiplicativeExpression n) {
      n.f0.accept(this);
      visitOperatorList(n.f1);
   }

   //
   // f0 -> ( "+"| "-"  ) UnaryExpression()
   //       | PreIncrementExpression()
   //       | PreDecrementExpression()
   //       | UnaryExpressionNotPlusMinus()
   //
   public void visit(UnaryExpression n) {
      n.f0.accept(this);
   }

   //
   // f0 -> "++"
   // f1 -> PrimaryExpression()
   //
   public void visit(PreIncrementExpression n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   //
   // f0 -> "--"
   // f1 -> PrimaryExpression()
   //
   public void visit(PreDecrementExpression n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   //
   // f0 -> ( "~"| "!"  ) UnaryExpression()
   //       | CastExpression()
   //       | PostfixExpression()
   //
   public void visit(UnaryExpressionNotPlusMinus n) {
      n.f0.accept(this);
   }

   //
   // f0 -> "(" PrimitiveType()
   //       | "(" Name() "[" "]"
   //       | "(" Name() ")" ( "~" | "!" | "(" | < IDENTIFIER >  | "this" | "super" | "new" | Literal()  )
   //
   public void visit(CastLookahead n) {
      n.f0.accept(this);
   }

   //
   // f0 -> PrimaryExpression()
   // f1 -> [ "++"| "--"  ]
   //
   public void visit(PostfixExpression n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   //
   // f0 -> "(" Type() ")" UnaryExpression()
   //       | "(" Type() ")" UnaryExpressionNotPlusMinus()
   //
   public void visit(CastExpression n) {
      n.f0.accept(this);
   }

   //
   // f0 -> PrimaryPrefix()
   // f1 -> ( PrimarySuffix() )*
   //
   public void visit(PrimaryExpression n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   //
   // f0 -> Literal()
   //       | Name()
   //       | "this"
   //       | "super" "." < IDENTIFIER >
   //       | "(" Expression() ")"
   //       | AllocationExpression()
   //
   public void visit(PrimaryPrefix n) {
      n.f0.accept(this);
   }

   //
   // f0 -> "." "this"
   //       | "." "class"
   //       | "." AllocationExpression()
   //       | "[" Expression() "]"
   //       | "." < IDENTIFIER >
   //       | Arguments()
   //
   public void visit(PrimarySuffix n) {
      n.f0.accept(this);
   }

   //
   // f0 -> < INTEGER_LITERAL >
   //       | < FLOATING_POINT_LITERAL >
   //       | < CHARACTER_LITERAL >
   //       | < STRING_LITERAL >
   //       | BooleanLiteral()
   //       | NullLiteral()
   //
   public void visit(Literal n) {
      n.f0.accept(this);
   }

   //
   // f0 -> "true"
   //       | "false"
   //
   public void visit(BooleanLiteral n) {
      n.f0.accept(this);
   }

   //
   // f0 -> "null"
   //
   public void visit(NullLiteral n) {
      n.f0.accept(this);
   }

   //
   // f0 -> "("
   // f1 -> [ ArgumentList() ]
   // f2 -> ")"
   //
   public void visit(Arguments n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
   }


    // f1 -> ( "," Expression() )*
    public void visitCommaSeparatedList(Enumeration e) {
	while (e.hasMoreElements()) {
	    visit((NodeSequence) e.nextElement(), " ");
	}
    }
	
    public void visitOperatorList(NodeListOptional n) {
	if ( n.present() ) {
	    for ( Enumeration e = n.elements(); e.hasMoreElements(); ) {
		space();
		visit((NodeSequence)e.nextElement(), " ");
	    }
	}
    }
    
   //
   // f0 -> Expression()
   // f1 -> ( "," Expression() )*
   //
   public void visit(ArgumentList n) {
      n.f0.accept(this);
      visitCommaSeparatedList(n.f1.elements());
   }

   //
   // f0 -> "new" PrimitiveType() ArrayDimensions() [ ArrayInitializer() ]
   //       | "new" Name() ( ArrayDimensions() [ ArrayInitializer() ]| Arguments() [ ClassBody() ]  )
   //
   public void visit(AllocationExpression n) {
//      visit(n.f0.choice, " ");
      // This code may break if "which" is removed from NodeChoice
      NodeSequence seq = (NodeSequence)n.f0.choice;

      if ( n.f0.which == 0 ) {
         seq.elementAt(0).accept(this);
         space();
         seq.elementAt(1).accept(this);
         seq.elementAt(2).accept(this);
         if ( ((NodeOptional)seq.elementAt(3)).present() ) {
            space();
            seq.elementAt(3).accept(this);
         }
      }
      else {   // needs a little tweaking
         seq.elementAt(0).accept(this);
         space();
         seq.elementAt(1).accept(this);

	 if (((NodeChoice) seq.elementAt(2)).which == 0) {
	     seq.elementAt(2).accept(this);
	 } else {     // Anonymous class
	     seq = (NodeSequence) ((NodeChoice) seq.elementAt(2)).choice;
	     seq.elementAt(0).accept(this);
	     if (((NodeOptional) seq.elementAt(1)).present()) {
		 space();
		 updateAnonymousClassLevel(+1);
		 seq.elementAt(1).accept(this);
		 updateAnonymousClassLevel(-1);
	     }
	 }
      }
   }

   //
   // f0 -> ( "[" Expression() "]" )+
   // f1 -> ( "[" "]" )*
   //
   public void visit(ArrayDimensions n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   //
   // f0 -> LabeledStatement()
   //       | Block()
   //       | EmptyStatement()
   //       | StatementExpression() ";"
   //       | SwitchStatement()
   //       | IfStatement()
   //       | WhileStatement()
   //       | DoStatement()
   //       | ForStatement()
   //       | BreakStatement()
   //       | ContinueStatement()
   //       | ReturnStatement()
   //       | ThrowStatement()
   //       | SynchronizedStatement()
   //       | TryStatement()
   //
   public void visit(Statement n) {
      n.f0.accept(this);
   }

   //
   // f0 -> < IDENTIFIER >
   // f1 -> ":"
   // f2 -> Statement()
   //
   public void visit(LabeledStatement n) {
      visit(n.f0); visit(n.f1); space();
      n.f2.accept(this);
   }

   //
   // f0 -> "{"
   // f1 -> ( BlockStatement() )*
   // f2 -> "}"
   //
   public void visit(Block n) {
      visit(n.f0);
      if ( n.f1.present() ) {
	  newLine();
         spc.updateSpc(+1);
         out.print(spc.spc);
         visit(n.f1, "\n" + spc.spc);
	 newLine();
         spc.updateSpc(-1);
	 out.print(spc.spc); visit(n.f2);
      } else
	  visit(n.f2);
   }

   //
   // f0 -> LocalVariableDeclaration() ";"
   //       | Statement()
   //       | UnmodifiedClassDeclaration()
   //
   public void visit(BlockStatement n) {
      n.f0.accept(this);
   }

   //
   // f0 -> [ "final" ]
   // f1 -> Type()
   // f2 -> VariableDeclarator()
   // f3 -> ( "," VariableDeclarator() )*
   //
   public void visit(LocalVariableDeclaration n) {
      if (n.f0.present()) {
	  visit(n.f0); space();
      }
      n.f1.accept(this);
      space();
      n.f2.accept(this);
      n.f3.accept(this);
   }

   //
   // f0 -> ";"
   //
   public void visit(EmptyStatement n) {
      n.f0.accept(this);
   }

   //
   // f0 -> PreIncrementExpression()
   //       | PreDecrementExpression()
   //       | Assignment()
   //       | PostfixExpression()
   //
   public void visit(StatementExpression n) {
      n.f0.accept(this);
   }

   //
   // f0 -> "switch"
   // f1 -> "("
   // f2 -> Expression()
   // f3 -> ")"
   // f4 -> "{"
   // f5 -> ( SwitchLabel() ( BlockStatement() )* )*
   // f6 -> "}"
   //
   public void visit(SwitchStatement n) {
      visit(n.f0); space(); visit(n.f1);
      n.f2.accept(this);
      visit(n.f3);
      space(); visit(n.f4); newLine();
      spc.updateSpc(+1);
      out.print(spc.spc);
      for ( Enumeration e = n.f5.elements(); e.hasMoreElements(); ) {
         NodeSequence seq = (NodeSequence)e.nextElement();
         seq.elementAt(0).accept(this);
	 space();
	 NodeListOptional list = (NodeListOptional) seq.elementAt(1);
         if (list.present() ) {
	     spc.updateSpc(+1);
	     visit(list, "\n" + spc.spc);
	     if (e.hasMoreElements()) {
		 NodeChoice node = (NodeChoice)
		     ((BlockStatement) list.elementAt(list.size() - 1)).f0;
		 if (node.which == 1) {
		     switch (((NodeChoice) ((Statement) node.choice).f0).which){
		     case 9: case 10: case 11: break;
		     default: newLine();
			 out.print(spc.spc + "// fall through");
		     }
		 }
	     }
	     newLine();
	     spc.updateSpc(-1);
	     if (e.hasMoreElements())
		 out.print(spc.spc);
         } else if (!e.hasMoreElements()) {
	     newLine();
	 }
      }
      spc.updateSpc(-1);
      out.print(spc.spc); visit(n.f6);
   }

   //
   // f0 -> "case" Expression() ":"
   //       | "default" ":"
   //
   public void visit(SwitchLabel n) {
      visit(n.f0.choice, "");
   }

   //
   // f0 -> "if"
   // f1 -> "("
   // f2 -> Expression()
   // f3 -> ")"
   // f4 -> Statement()
   // f5 -> [ "else" Statement() ]
   //
   public void visit(IfStatement n) {
      visit(n.f0); space(); visit(n.f1);
      n.f2.accept(this);
      visit(n.f3);

      boolean isIfBlock = (n.f4.f0.which == 1);
      
      if ( !isIfBlock ) {    // only indent if it's not a Block
         spc.updateSpc(+1);
	 newLine();
	 out.print(spc.spc);
      } else {
	  space();
      }
      
      n.f4.accept(this);
      
      if ( n.f4.f0.which != 1 )
         spc.updateSpc(-1);

      if ( n.f5.present() ) {
	  if (!isIfBlock) {
	     newLine();
	     out.print(spc.spc);
	  } else {
	      space();
	  }

	  int kindOfElseBlock = ((Statement)((NodeSequence)n.f5.node)
				.elementAt(1)).f0.which;
	  
	  if (kindOfElseBlock != 1 && kindOfElseBlock != 5) {
            spc.updateSpc(+1);

	    visit(n.f5.node, "\n" + spc.spc);

            spc.updateSpc(-1);
	  } else {
	      visit(n.f5.node, " ");
	  }
      }
   }

   //
   // f0 -> "while"
   // f1 -> "("
   // f2 -> Expression()
   // f3 -> ")"
   // f4 -> Statement()
   //
    public void visit(WhileStatement n) {
	visit(n.f0); space(); visit(n.f1);
	n.f2.accept(this);
	visit(n.f3);

	if ( n.f4.f0.which != 1 ) {
	    newLine();
	    spc.updateSpc(+1);
	    out.print(spc.spc);
	    n.f4.accept(this);
	    spc.updateSpc(-1);
	} else {
	    space();
	    n.f4.accept(this);
	}
    }

   //
   // f0 -> "do"
   // f1 -> Statement()
   // f2 -> "while"
   // f3 -> "("
   // f4 -> Expression()
   // f5 -> ")"
   // f6 -> ";"
   //
    public void visit(DoStatement n) {
	visit(n.f0);

	if ( n.f1.f0.which != 1 ) {
	    newLine();
	    spc.updateSpc(+1);
	    out.print(spc.spc);
	    n.f1.accept(this);
	    spc.updateSpc(-1);
	    newLine();
	    out.print(spc.spc);
	} else {
	    space();
	    n.f1.accept(this);
	    space();
	}
	visit(n.f2); space(); visit(n.f3);
	n.f4.accept(this);
	visit(n.f5); visit(n.f6);
    }

   //
   // f0 -> "for"
   // f1 -> "("
   // f2 -> [ ForInit() ]
   // f3 -> ";"
   // f4 -> [ Expression() ]
   // f5 -> ";"
   // f6 -> [ ForUpdate() ]
   // f7 -> ")"
   // f8 -> Statement()
   //
   public void visit(ForStatement n) {
      visit(n.f0); space(); visit(n.f1);
      n.f2.accept(this);
      visit(n.f3);
      if (n.f4.present())
	  space();
      n.f4.accept(this);
      visit(n.f5);
      if (n.f6.present())
	  space();
      n.f6.accept(this);
      visit(n.f7);
      
      if ( n.f8.f0.which != 1 ) {
	  newLine();
	  spc.updateSpc(+1);
	  out.print(spc.spc);
	  n.f8.accept(this);
	  spc.updateSpc(-1);
      } else {
	  space();
	  n.f8.accept(this);
      }
   }

   //
   // f0 -> LocalVariableDeclaration()
   //       | StatementExpressionList()
   //
   public void visit(ForInit n) {
      n.f0.accept(this);
   }

   //
   // f0 -> StatementExpression()
   // f1 -> ( "," StatementExpression() )*
   //
   public void visit(StatementExpressionList n) {
      n.f0.accept(this);
      visitCommaSeparatedList(n.f1.elements());
   }

   //
   // f0 -> StatementExpressionList()
   //
   public void visit(ForUpdate n) {
      n.f0.accept(this);
   }

   //
   // f0 -> "break"
   // f1 -> [ < IDENTIFIER >  ]
   // f2 -> ";"
   //
   public void visit(BreakStatement n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
	  space(); visit((Token) n.f1.node);
      }
      n.f2.accept(this);
   }

   //
   // f0 -> "continue"
   // f1 -> [ < IDENTIFIER >  ]
   // f2 -> ";"
   //
   public void visit(ContinueStatement n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
	  space();
	  visit((Token) n.f1.node);
      }
      n.f2.accept(this);
   }

   //
   // f0 -> "return"
   // f1 -> [ Expression() ]
   // f2 -> ";"
   //
   public void visit(ReturnStatement n) {
      visit(n.f0);
      if ( n.f1.present() ) {
         space();
         n.f1.accept(this);
      }
      n.f2.accept(this);
   }

   //
   // f0 -> "throw"
   // f1 -> Expression()
   // f2 -> ";"
   //
   public void visit(ThrowStatement n) {
      visit(n.f0); space();
      n.f1.accept(this);
      n.f2.accept(this);
   }

   //
   // f0 -> "synchronized"
   // f1 -> "("
   // f2 -> Expression()
   // f3 -> ")"
   // f4 -> Block()
   //
   public void visit(SynchronizedStatement n) {
      visit(n.f0); space(); visit(n.f1);
      n.f2.accept(this);
      visit(n.f3);
      
	  space();
	  n.f4.accept(this);
   }

   //
   // f0 -> "try"
   // f1 -> Block()
   // f2 -> ( "catch" "(" FormalParameter() ")" Block() )*
   // f3 -> [ "finally" Block() ]
   //
   public void visit(TryStatement n) {
      visit(n.f0); space();
      n.f1.accept(this);
      for ( Enumeration e = n.f2.elements(); e.hasMoreElements(); ) {
         NodeSequence seq = (NodeSequence)e.nextElement();
	 space();
         seq.elementAt(0).accept(this);
         space();
         seq.elementAt(1).accept(this);
         seq.elementAt(2).accept(this);
         seq.elementAt(3).accept(this);
         space();
         seq.elementAt(4).accept(this);
      }

      if ( n.f3.present() ) {
         NodeSequence seq = (NodeSequence)n.f3.node;
	 space();
         seq.elementAt(0).accept(this);
	 space();
         seq.elementAt(1).accept(this);
      }
   }
}

//
// For indentation
//
class Spacing {
   final int INDENT_AMT;

   String spc = "";
   int indentLevel = 0;

   public Spacing(int indentAmt) {
      INDENT_AMT = indentAmt;
   }

   public String toString() {
      return spc;
   }

   public void updateSpc(int numIndentLvls) {
      indentLevel += numIndentLvls;

      if ( numIndentLvls < 0 )
         spc = spc.substring(-1 * numIndentLvls * INDENT_AMT);
      else if ( numIndentLvls > 0 ) {
         StringBuffer buf = new StringBuffer(spc);

         for ( int i = 0; i < numIndentLvls * INDENT_AMT; ++i )
            buf.append(" ");

         spc = buf.toString();
      }
   }
}


/*
interface KindOfSeparator {
    public static final int SEQUENCE = 0;
    public static final int SEQUENCE_FINAL = 1;
    public static final int LIST = 2;
    public static final int LIST_FINAL = 3;
}

interface SeparationCosts {
    public static final int BEFORE_PUNCTUATION = 20;
    public static final int NONE = 0;
    public static final int ASSIGNMENT = 1;
    public static final int ARGUMENT_LIST = 3;
    public static final int ARGUMENT = 2;
    public static final int CONDITIONAL_EXPRESSION = 4;
    public static final int CONDITIONAL_AND_EXPRESSION = 5;
    public static final int INCLUSIVE_OR_EXPRESSION = 6;
    public static final int EXCLUSIVE_OR_EXPRESSION = 7;
    public static final int AND_EXPRESSION = 8;
    public static final int EQUALITY_EXPRESSION = 9;
    public static final int INSTANCEOF_EXPRESSION = 10;
    public static final int RELATIONAL_EXPRESSION = 11;
    public static final int SHIFT_EXPRESSION = 12;
    public static final int ADDITIVE_EXPRESSION = 13; 
    public static final int MULTIPLICATIVE_EXPRESSION = 14;
    public static final int UNARY_EXPRESSION = 15;
}

class Separator implements KindOfSeparator, SeparationCosts {
    String separators[] = {"", "", "", ""};
    int separationCosts[] = 
    
    public Separator() {}
}
*/
/* Hmm. Das ist kein gutes Design.
   Wie waer's damit, einen weiteren Stack zu verwenden, der die einzelnen
   Separators abspeichert? Dann koennte ich zwei auf den Stack legen und
   braeuchte nicht KindOfSeparator zu unterscheiden.

   Ein Attribut koennte die Klammerungstiefe abspeichern und damit die
   separationCosts fuer die Operatoren hochsetzen.
   Die Kosten fuer eine neue Zeile muessen kleiner sein als die der Operatoren,
   aber groesser als die BEFORE_PUNCTUATION (= default).

   Man koennte auch die separator in den Vektor eintragen und nur diese
   betrachten, vom Default (BEFORE_PUNCTUATION) abgesehen.

   Die Laenge des Separators wird bei der Berechnung des Umbruchs abgezogen. */
